
# pylint: disable=eval-used

import os
import sys
import time

import threading         as mt
import multiprocessing   as mp

import radical.utils     as ru

from .. import Session
from .. import utils     as rpu
from .. import constants as rpc


# ------------------------------------------------------------------------------
#
class Worker(rpu.Component):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session=None):

        self._session = session

        if isinstance(cfg, str): cfg = ru.Config(cfg=ru.read_json(cfg))
        else                   : cfg = ru.Config(cfg=cfg)


        # generate a MPI rank dependent UID for each worker process
        # FIXME: this should be delegated to ru.generate_id

        # FIXME: why do we need to import `os` again after MPI Spawn?
        import os                                                         # noqa

        # FIXME: rank determination should be moved to RU
        rank = None

        if rank is None: rank = os.environ.get('PMIX_RANK')
        if rank is None: rank = os.environ.get('PMI_RANK')
        if rank is None: rank = os.environ.get('OMPI_COMM_WORLD_RANK')

        if rank is not None:
            cfg['uid'] = '%s.%03d' % (cfg['uid'], int(rank))

        self._n_cores = cfg.cores
        self._n_gpus  = cfg.gpus

        self._info    = ru.Config(cfg=cfg.get('info', {}))


        if not self._session:
            self._session = Session(cfg=cfg, uid=cfg.sid, _primary=False)


        rpu.Component.__init__(self, cfg, self._session)

        self._term    = mp.Event()          # set to terminate
        self._res_evt = mp.Event()          # set on free resources

        self._mlock   = ru.Lock(self._uid)  # lock `_modes`
        self._modes   = dict()              # call modes (call, exec, eval, ...)

        # We need to make sure to run only up to `gpn` tasks using a gpu
        # within that pool, so need a separate counter for that.
        self._resources = {'cores' : [0] * self._n_cores,
                           'gpus'  : [0] * self._n_gpus}

        # resources are initially all free
        self._res_evt.set()

      # # create a multiprocessing pool with `cpn` worker processors.  Set
      # # `maxtasksperchild` to `1` so that we get a fresh process for each
      # # task.  That will also allow us to run command lines via `exec`,
      # # effectively replacing the worker process in the pool for a specific
      # # task.
      # #
      # # We use a `fork` context to inherit log and profile handles.
      # #
      # # NOTE: The mp documentation is wrong; mp.Pool does *not* have a context
      # #       parameters.  Instead, the Pool has to be created within
      # #       a context.
      # ctx = mp.get_context('fork')
      # self._pool = ctx.Pool(processes=self._n_cores,
      #                       initializer=None,
      #                       maxtasksperchild=1)
      # NOTE: a multiprocessing pool won't work, as pickle is not able to
      #       serialize our worker object.  So we use our own process pool.
      #       It's not much of a loss since we want to respawn new processes for
      #       each task anyway (to improve isolation).
        self._pool  = dict()  # map task uid to process instance
        self._plock = ru.Lock('p' + self._uid)  # lock _pool

        # We also create a queue for communicating results back, and a thread to
        # watch that queue
        self._result_queue = mp.Queue()
        self._result_thead = mt.Thread(target=self._result_watcher)
        self._result_thead.daemon = True
        self._result_thead.start()

        # connect to master
        self.register_subscriber(rpc.CONTROL_PUBSUB, self._control_cb)
        self.register_publisher(rpc.CONTROL_PUBSUB)

        # run worker initialization *before* starting to work on requests.
        # the worker provides three builtin methods:
        #     eval:  evaluate a piece of python code
        #     exec:  execute  a command line (fork/exec)
        #     shell: execute  a shell command
        #     call:  execute  a method or function call
        self.register_mode('call',  self._call)
        self.register_mode('eval',  self._eval)
        self.register_mode('exec',  self._exec)
        self.register_mode('shell', self._shell)

        self.pre_exec()

        # connect to the request / response ZMQ queues
        self._res_put = ru.zmq.Putter('to_res', self._info.res_addr_put)
        self._req_get = ru.zmq.Getter('to_req', self._info.req_addr_get,
                                                cb=self._request_cb)

        # the worker can return custom information which will be made available
        # to the master.  This can be used to communicate, for example, worker
        # specific communication endpoints.

        # `info` is a placeholder for any additional meta data communicated to
        # the worker
        self.publish(rpc.CONTROL_PUBSUB, {'cmd': 'worker_register',
                                          'arg': {'uid' : self._uid,
                                                  'info': self._info}})


    # --------------------------------------------------------------------------
    #
    # This class-method creates the appropriate sub-class for the Stager
    #
    @classmethod
    def create(cls, cfg, session):

        return Worker(cfg, session)


    # --------------------------------------------------------------------------
    #
    def pre_exec(self):
        '''
        This method can be overloaded by the Worker implementation to run any
        pre_exec commands before spawning worker processes.
        '''

        pass


    # --------------------------------------------------------------------------
    #
    def register_mode(self, name, executor):

        assert(name not in self._modes)

        self._modes[name] = executor


    # --------------------------------------------------------------------------
    #
    def _call(self, data):
        '''
        We expect data to have a three entries: 'method' or 'function',
        containing the name of the member method or the name of a free function
        to call, `args`, an optional list of unnamed parameters, and `kwargs`,
        and optional dictionary of named parameters.
        '''

        if 'method' in data:
            to_call = getattr(self, data['method'], None)

        elif 'function' in data:
            names   = dict(list(globals().items()) + list(locals().items()))
            to_call = names.get(data['function'])

        else:
            raise ValueError('no method or function specified: %s' % data)

        if not to_call:
            raise ValueError('callable not found: %s' % data)


        args   = data.get('args',   [])
        kwargs = data.get('kwargs', {})

        try:
            out = to_call(*args, **kwargs)
            err = None
            ret = 0

        except Exception as e:
            self._log.exception('_call failed: %s' % (data))
            out = None
            err = 'call failed: %s' % e
            ret = 1

        return out, err, ret


    # --------------------------------------------------------------------------
    #
    # FIXME: an MPI call mode should be added.  That could work along these
    #        lines of:
    #
    # --------------------------------------------------------------------------
    #  def _mpi(self, data):
    #
    #      try:
    #          cmd = rp.agent.launch_method.construct_command(data,
    #                  executable=self.exe, args=data['func'])
    #          out = rp.sh_callout(cmd)
    #          err = None
    #          ret = 0
    #
    #      except Exception as e:
    #          self._log.exception('_mpi failed: %s' % (data))
    #          out = None
    #          err = 'mpi failed: %s' % e
    #          ret = 1
    #
    #      return out, err, ret
    # --------------------------------------------------------------------------
    #
    # For that to work we would need to be able to create a LM here, but ideally
    # not replicate the work done in the agent executor.


    # --------------------------------------------------------------------------
    #
    def _eval(self, data):
        '''
        We expect data to have a single entry: 'code', containing the Python
        code to be eval'ed
        '''

        try:
            out = eval(data['code'])
            err = None
            ret = 0

        except Exception as e:
            self._log.exception('_eval failed: %s' % (data))
            out = None
            err = 'eval failed: %s' % e
            ret = 1

        return out, err, ret


    # --------------------------------------------------------------------------
    #
    def _exec(self, data):
        '''
        We expect data to have two entries: 'exe', containing the executabele to
        run, and `args` containing a list of arguments (strings) to pass as
        command line arguments.  We use `sp.Popen` to run the fork/exec, and to
        collect stdout, stderr and return code
        '''

        try:
            import subprocess as sp

            exe  = data['exe']
            args = data.get('args', list())
            env  = data.get('env',  dict())

            args = '%s %s' % (exe, ' '.join(args))

            proc = sp.Popen(args=args,      env=env,
                            stdin=None,     stdout=sp.PIPE, stderr=sp.PIPE,
                            close_fds=True, shell=True)
            out, err = proc.communicate()
            ret      = proc.returncode

        except Exception as e:
            self._log.exception('_exec failed: %s' % (data))
            out = None
            err = 'exec failed: %s' % e
            ret = 1

        return out, err, ret


    # --------------------------------------------------------------------------
    #
    def _shell(self, data):
        '''
        We expect data to have a single entry: 'cmd', containing the command
        line to be called as string.
        '''

        try:
            out, err, ret = ru.sh_callout(data['cmd'], shell=True)

        except Exception as e:
            self._log.exception('_shell failed: %s' % (data))
            out = None
            err = 'shell failed: %s' % e
            ret = 1

        return out, err, ret


    # --------------------------------------------------------------------------
    #
    def _alloc_task(self, task):
        '''
        allocate task resources
        '''

        with self._mlock:

            cores = task.get('cores', 1)
            gpus  = task.get('gpus' , 0)

            assert(cores >= 1)
            assert(cores <= self._n_cores)
            assert(gpus  <= self._n_gpus)

            if cores > self._resources['cores'].count(0): return False
            if gpus  > self._resources['gpus' ].count(0): return False

            alloc_cores = list()
            alloc_gpus  = list()

            if cores:
                for n in range(self._n_cores):
                    if not self._resources['cores'][n]:
                        self._resources['cores'][n] = 1
                        alloc_cores.append(n)
                        if len(alloc_cores) == cores:
                            break

            if gpus:
                for n in range(self._n_gpus):
                    if not self._resources['gpus'][n]:
                        self._resources['gpus'][n] = 1
                        alloc_gpus.append(n)
                        if len(alloc_gpus) == gpus:
                            break

            task['resources'] = {'cores': alloc_cores,
                                 'gpus' : alloc_gpus}
            return True


    # --------------------------------------------------------------------------
    #
    def _dealloc_task(self, task):
        '''
        deallocate task resources
        '''

        with self._mlock:

            resources = task['resources']

            for n in resources['cores']:
                assert(self._resources['cores'][n])
                self._resources['cores'][n] = 0

            for n in resources['gpus']:
                assert(self._resources['gpus'][n])
                self._resources['gpus'][n] = 0

            # signal available resources
            self._res_evt.set()

            return True


    # --------------------------------------------------------------------------
    #
    def _request_cb(self, tasks):
        '''
        grep call type from tasks, check if methods are registered, and
        invoke them.
        '''

        for task in ru.as_list(tasks):

            self._prof.prof('reg_start', uid=self._uid, msg=task['uid'])
            task['worker'] = self._uid

            try:
                # ok, we have work to do.  Check the requirements to see how
                # many cpus and gpus we need to mark as busy
                while not self._alloc_task(task):
                    # no resource - wait for new resources
                    #
                    # NOTE: this will block smaller tasks from being executed
                    #       right now.  alloc_task is not a proper scheduler,
                    #       after all.
                    while not self._res_evt.wait(timeout=1.0):

                        # break on termination
                        if self._term.is_set():
                            return False

                    self._res_evt.clear()

                # we got an allocation for this task, and can run it, so apply
                # to the process pool.  The callback (`self._result_cb`) will
                # pick the task up on completion and free resources.
                #
                # NOTE: we don't use mp.Pool - see __init__ for details

              # ret = self._pool.apply_async(func=self._dispatch, args=[task],
              #                              callback=self._result_cb,
              #                              error_callback=self._error_cb)
                proc = mp.Process(target=self._dispatch, args=[task],
                                  daemon=True)

                with self._plock:

                    # we need to include `proc.start()` in the lock, as
                    # otherwise we may end up getting the `self._result_cb`
                    # before the pid could be registered in `self._pool`.
                    proc.start()
                    self._pool[proc.pid] = proc
                    self._log.debug('applied: %s: %s: %s',
                                    task['uid'], proc.pid, self._pool.keys())

            except Exception as e:

                self._log.exception('request failed')

                # free resources again for failed task
                self._dealloc_task(task)

                res = {'req': task['uid'],
                       'out': None,
                       'err': 'req_cb error: %s' % e,
                       'ret': 1}

                self._res_put.put(res)


    # --------------------------------------------------------------------------
    #
    def _dispatch(self, task):

        # this method is running in a process of the process pool, and will now
        # apply the task to the respective execution mode.
        #
        # NOTE: application of pre_exec directives may got here

        task['pid'] = os.getpid()

        # ----------------------------------------------------------------------
        def _dispatch_thread(tlock):
            out, err, ret = self._modes[mode](task.get('data'))
            with tlock:
                res = [task, str(out), str(err), int(ret)]
                self._log.debug('put 1 result: task %s', task['uid'])
                self._result_queue.put(res)
        # ----------------------------------------------------------------------


        ret = None
        try:
          # self._log.debug('dispatch: %s: %d', task['uid'], task['pid'])
            mode = task['mode']
            assert(mode in self._modes), 'no such call mode %s' % mode

            tout = task.get('timeout')
            self._log.debug('dispatch with tout %s', tout)

            tlock  = mt.Lock()
            thread = mt.Thread(target=_dispatch_thread,
                               args=[tlock])
            thread.daemon = True
            thread.start()
            thread.join(timeout=tout)

            with tlock:
                if thread.is_alive():
                    out = None
                    err = 'timeout (>%s)' % tout
                    ret = 1
                    res = [task, str(out), str(err), int(ret)]
                    self._log.debug('put 2 result: task %s', task['uid'])
                    self._result_queue.put(res)

          # self._log.debug('dispatch done: %s', task['uid'])

        except Exception as e:

            self._log.exception('dispatch failed')
            out = None
            err = 'dispatch failed: %s' % e
            ret = 1
            res = [task, str(out), str(err), int(ret)]
            self._log.debug('put 3 result: task %s', task['uid'])
            self._result_queue.put(res)

        finally:
            # if we kill the process too quickly, the result put above
            # will not make it out, thus make sure the queue is empty
            # first.
            ret = 1
            self._result_queue.close()
            self._result_queue.join_thread()
            sys.exit(ret)
          # os.kill(os.getpid(), signal.SIGTERM)



    # --------------------------------------------------------------------------
    #
    def _result_watcher(self):

        while True:

            try:
                res = self._result_queue.get()
                self._log.debug('got   result: %s', res)
                self._result_cb(res)
            except:
                self._log.exception('queue error')
                raise


    # --------------------------------------------------------------------------
    #
    def _result_cb(self, result):

        try:
            task, out, err, ret = result
          # self._log.debug('result cb: task %s', task['uid'])

            with self._plock:
                pid  = task['pid']
                del(self._pool[pid])

            # free resources again for the task
            self._dealloc_task(task)

            res = {'req': task['uid'],
                   'out': out,
                   'err': err,
                   'ret': ret}

            self._res_put.put(res)
            self._prof.prof('reg_stop', uid=self._uid, msg=task['uid'])
        except:
            self._log.exception('result cb failed')
            raise



    # --------------------------------------------------------------------------
    #
    def _error_cb(self, error):

        self._log.debug('error: %s', error)
        raise RuntimeError(error)


    # --------------------------------------------------------------------------
    #
    def _control_cb(self, topic, msg):

        if msg['cmd'] == 'worker_terminate':
            if msg['arg']['uid'] == self._uid:

                self._log.debug('got terminate msg: %s: %s', topic, msg)

                self._term.set()
                self.stop()
                sys.exit(0)


    # --------------------------------------------------------------------------
    #
    def run(self):

        while not self._term.is_set():
            time.sleep(1)


# ------------------------------------------------------------------------------




__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import radical.utils as ru

from .base import LaunchMethod


# ------------------------------------------------------------------------------
#
class MPIRun_MPT(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, cfg, session):

        LaunchMethod.__init__(self, name, cfg, session)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        self.launch_command = ru.which([
            'mpirun',            # General case
            'mpirun_rsh',        # Gordon @ SDSC
            'mpirun-mpich-mp',   # Mac OSX MacPorts
            'mpirun-openmpi-mp'  # Mac OSX MacPorts
        ])

        self.mpi_version, self.mpi_flavor = self._get_mpi_info(self.launch_command)


    # --------------------------------------------------------------------------
    #
    def construct_command(self, cu, launch_script_hop):

        slots        = cu['slots']
        cud          = cu['description']
        task_exec    = cud['executable']
        task_env     = cud.get('environment') or dict()
        task_args    = cud.get('arguments')   or list()
        task_argstr  = self._create_arg_string(task_args)

        # Construct the executable and arguments
        if task_argstr: task_command = "%s %s" % (task_exec, task_argstr)
        else          : task_command = task_exec

        env_string = ''
        env_list   = self.EXPORT_ENV_VARIABLES + list(task_env.keys())
        if env_list:

            if self.mpi_flavor == self.MPI_FLAVOR_HYDRA:
                env_string = '-envlist "%s"' % ','.join(env_list)

            elif self.mpi_flavor == self.MPI_FLAVOR_OMPI:
                for var in env_list:
                    env_string += '-x "%s" ' % var


        if 'nodes' not in slots:
            raise RuntimeError('insufficient information to launch via %s: %s'
                              % (self.name, slots))

        # Extract all the hosts from the slots
        hostlist = list()
        for node in slots['nodes']:
            for cpu_proc in node['core_map']:
                hostlist.append(node['name'])
            for gpu_proc in node['gpu_map']:
                hostlist.append(node['name'])
        hosts_string = ",".join(hostlist)

        # On Cheyenne (which is the only machine that requires mpirun_mpt
        # currently), we also have to set MPI_SHEPHERD=true, we add it to the
        # environment of the cu which will get parsed in the spawner
        if 'environment' not in list(cu['description'].keys()):
            cu['description']['environment'] = dict()
        cu['description']['environment']['MPI_SHEPHERD'] = True

        # -np:  usually it is len(hostlist), as it means run N processes over
        # N hosts but for Cheyenne it is 1 since specification of the host means
        # land N processes on EACH host, where N is specified as arg to -np
        command = "%s %s -np %d %s %s" % (self.launch_command, hosts_string, 1,
                                          env_string, task_command)

        return command, None


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import time
import math

from ... import constants as rpc

from .base import AgentSchedulingComponent


# ------------------------------------------------------------------------------
#
class Torus(AgentSchedulingComponent):

    # TODO: Ultimately all BG/Q specifics should move out of the scheduler

    # --------------------------------------------------------------------------
    #
    # Offsets into block structure
    #
    TORUS_BLOCK_INDEX  = 0
    TORUS_BLOCK_COOR   = 1
    TORUS_BLOCK_NAME   = 2
    TORUS_BLOCK_STATUS = 3


    # --------------------------------------------------------------------------
    def __init__(self, cfg, session):

        self.nodes            = None
        self._cores_per_node  = None

        AgentSchedulingComponent.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def _configure(self):
        if not self._rm_cores_per_node:
            raise RuntimeError("ResourceManager %s didn't _configure cores_per_node." %
                               self._rm_info['name'])

        self._cores_per_node = self._rm_cores_per_node

        # TODO: get rid of field below
        self.nodes = 'bogus'


    # --------------------------------------------------------------------------
    #
    def slot_status(self):
        """Returns a multi-line string corresponding to slot status.
        """

        slot_matrix = ""
        for slot in self._rm.torus_block:
            slot_matrix += "|"
            if slot[self.TORUS_BLOCK_STATUS] == rpc.FREE:
                slot_matrix += "-" * self._rm_cores_per_node
            else:
                slot_matrix += "+" * self._rm_cores_per_node
        slot_matrix += "|"
        return {'timestamp': time.time(),
                'slotstate': slot_matrix}


    # --------------------------------------------------------------------------
    #
    # Allocate a number of cores
    #
    # Currently only implements full-node allocation, so core count must
    # be a multiple of cores_per_node.
    #
    def _allocate_slot(self, cores_requested, gpus_requested):

        block = self._rm.torus_block
        sub_block_shape_table = self._rm.shape_table

        self._log.info("Trying to allocate %d core(s / %d gpus.",
                cores_requested, gpus_requested)

        # FIXME GPU

        if cores_requested % self._rm_cores_per_node:
            num_cores = int(math.ceil(cores_requested / float(self._rm_cores_per_node))) \
                        * self._rm_cores_per_node
            self._log.error('Core not multiple of %d, increasing to %d!',
                           self._rm_cores_per_node, num_cores)

        num_nodes = cores_requested / self._rm_cores_per_node

        offset = self._alloc_sub_block(block, num_nodes)

        if offset is None:
            self._log.warning('No allocation made.')
            return

        # TODO: return something else than corner location? Corner index?
        sub_block_shape     = sub_block_shape_table[num_nodes]
        sub_block_shape_str = self._rm.shape2str(sub_block_shape)
        corner              = block[offset][self.TORUS_BLOCK_COOR]
        corner_offset       = self.corner2offset(self._rm.torus_block, corner)
        corner_node         = self._rm.torus_block[corner_offset][self.TORUS_BLOCK_NAME]

        end = self.get_last_node(corner, sub_block_shape)
        self._log.debug('Allocating sub-block of %d node(s) with dimensions %s'
                       ' at offset %d with corner %s and end %s.',
                        num_nodes, sub_block_shape_str, offset,
                        self._rm.loc2str(corner), self._rm.loc2str(end))

        return {'cores_per_node'      : self._rm_cores_per_node,
                'loadl_bg_block'      : self._rm.loadl_bg_block,
                'sub_block_shape_str' : sub_block_shape_str,
                'corner_node'         : corner_node,
                'lm_info'             : self._rm_lm_info}


    # --------------------------------------------------------------------------
    #
    # Allocate a sub-block within a block
    # Currently only works with offset that are exactly the sub-block size
    #
    def _alloc_sub_block(self, block, num_nodes):

        offset = 0
        # Iterate through all nodes with offset a multiple of the sub-block size
        while True:

            # Verify the assumption (needs to be an assert?)
            if offset % num_nodes != 0:
                msg = 'Sub-block needs to start at correct offset!'
                self._log.exception(msg)
                raise ValueError(msg)
                # TODO: If we want to workaround this, the coordinates need to overflow

            not_free = False
            # Check if all nodes from offset till offset+size are rpc.FREE
            for peek in range(num_nodes):
                try:
                    if block[offset + peek][self.TORUS_BLOCK_STATUS] \
                                                                    == rpc.BUSY:
                        # Once we find the first rpc.BUSY node we
                        # can discard this attempt
                        not_free = True
                        break
                except IndexError:
                    self._log.exception('Block out of bound. Num_nodes: %d,'
                                        'offset: %d, peek: %d.',
                                        num_nodes, offset, peek)

            if not_free is True:
                # No success at this offset
                self._log.info("No free nodes found at this offset: %d.", offset)

                # If we weren't the last attempt, then increase the offset and iterate again.
                if offset + num_nodes < self._block2num_nodes(block):
                    offset += num_nodes
                    continue
                else:
                    return

            else:
                # At this stage we have found a free spot!

                self._log.info("Free nodes found at this offset: %d.", offset)

                # Then mark the nodes busy
                for peek in range(num_nodes):
                    block[offset + peek][self.TORUS_BLOCK_STATUS] = rpc.BUSY

                return offset


    # --------------------------------------------------------------------------
    #
    # Return the number of nodes in a block
    #
    def _block2num_nodes(self, block):
        return len(block)


    # --------------------------------------------------------------------------
    #
    def _release_slot(self, xxx_todo_changeme):
        (corner, shape) = xxx_todo_changeme
        self._free_cores(self._rm.torus_block, corner, shape)


    # --------------------------------------------------------------------------
    #
    # Free up an allocation
    #
    def _free_cores(self, block, corner, shape):

        # Number of nodes to free
        num_nodes = self._shape2num_nodes(shape)

        # Location of where to start freeing
        offset = self.corner2offset(block, corner)

        self._log.info("Freeing %d nodes starting at %d.", num_nodes, offset)

        for peek in range(num_nodes):
            assert block[offset + peek][self.TORUS_BLOCK_STATUS] == rpc.BUSY, \
                   'Block %d not Free!' % block[offset + peek]
            block[offset + peek][self.TORUS_BLOCK_STATUS] = rpc.FREE


    # --------------------------------------------------------------------------
    #
    # Follow coordinates to get the last node
    #
    def get_last_node(self, origin, shape):

        ret = dict()

        for dim in self._rm.torus_dimension_labels:
            ret[dim] = origin[dim] + shape[dim] - 1
        return ret


    # --------------------------------------------------------------------------
    #
    # Return the number of nodes for the given block shape
    #
    def _shape2num_nodes(self, shape):

        nodes = 1
        for dim in self._rm.torus_dimension_labels:
            nodes *= shape[dim]

        return nodes


    # --------------------------------------------------------------------------
    #
    # Return the offset into the node list from a corner
    #
    # TODO: Can this be determined instead of searched?
    #
    def corner2offset(self, block, corner):
        offset = 0

        for e in block:
            if corner == e[self.TORUS_BLOCK_COOR]:
                return offset
            offset += 1

        return offset


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import time
import threading     as mt
import subprocess    as mp
import radical.utils as ru

from .base import LaunchMethod


# ------------------------------------------------------------------------------
#
# NOTE: This requires a development version of Open MPI available.
#
class ORTE(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, cfg, session):

        LaunchMethod.__init__(self, name, cfg, session)

        # We remove all ORTE related environment variables from the launcher
        # environment, so that we can use ORTE for both launch of the
        # (sub-)agent and CU execution.
        self.env_removables.extend(["OMPI_", "OPAL_", "PMIX_"])


    # --------------------------------------------------------------------------
    #
    @classmethod
    def rm_config_hook(cls, name, cfg, rm, log, profiler):
        """
        FIXME: this config hook will manipulate the ResourceManager nodelist.
               Not a nice thing to do, but hey... :P What really should be
               happening is that the ResourceManager digs information on node
               reservation out of the config and configures the node list
               accordingly.  This config hook should be limited to starting the
               DVM.
        """

        dvm_command = ru.which('orte-dvm')
        if not dvm_command:
            raise Exception("Couldn't find orte-dvm")

        # Now that we found the orte-dvm, get ORTE version
        out, _, _ = ru.sh_callout('orte-info | grep "Open RTE"', shell=True)
        orte_info = dict()
        for line in out.split('\n'):

            line = line.strip()
            if not line:
                continue

            key, val = line.split(':', 1)
            if 'Open RTE' == key.strip():
                orte_info['version'] = val.strip()
            elif  'Open RTE repo revision' == key.strip():
                orte_info['version_detail'] = val.strip()

        assert(orte_info.get('version'))
        log.info("Found Open RTE: %s / %s",
                 orte_info['version'], orte_info.get('version_detail'))

        # Use (g)stdbuf to disable buffering.
        # We need this to get the "DVM ready",
        # without waiting for orte-dvm to complete.
        # The command seems to be generally available on our Cray's,
        # if not, we can code some home-coooked pty stuff.
        stdbuf_cmd =  ru.which(['stdbuf', 'gstdbuf'])
        if not stdbuf_cmd:
            raise Exception("Couldn't find (g)stdbuf")
        stdbuf_arg = "-oL"

        # Base command = (g)stdbuf <args> + orte-dvm + dvm-args + debug_args
        dvm_args = '--report-uri -'
        cmdline  = '%s %s %s %s ' % (stdbuf_cmd, stdbuf_arg, dvm_command, dvm_args)

        # Additional (debug) arguments to orte-dvm
        if os.environ.get('RADICAL_PILOT_ORTE_VERBOSE'):
            debug_strings = [
                             '--debug-devel',
                             '--mca odls_base_verbose 100',
                             '--mca rml_base_verbose 100'
                            ]
        else:
            debug_strings = []

        cmdline += ' '.join(debug_strings)

        vm_size = len(rm.node_list)
        log.info("Start DVM on %d nodes ['%s']", vm_size, ' '.join(cmdline))
        profiler.prof(event='dvm_start', uid=cfg['pid'])

        dvm_uri     = None
        dvm_process = mp.Popen(cmdline.split(), stdout=mp.PIPE, stderr=mp.STDOUT)

        while True:

            line = dvm_process.stdout.readline().strip()

            if line.startswith('VMURI:'):

                if len(line.split(' ')) != 2:
                    raise Exception("Unknown VMURI format: %s" % line)

                label, dvm_uri = line.split(' ', 1)

                if label != 'VMURI:':
                    raise Exception("Unknown VMURI format: %s" % line)

                log.info("ORTE DVM URI: %s" % dvm_uri)

            elif line == 'DVM ready':

                if not dvm_uri:
                    raise Exception("VMURI not found!")

                log.info("ORTE DVM startup successful!")
                profiler.prof(event='dvm_ok', uid=cfg['pid'])
                break

            else:

                # Check if the process is still around,
                # and log output in debug mode.
                if dvm_process.poll() is None:
                    log.debug("ORTE: %s", line)
                else:
                    # Process is gone: fatal!
                    profiler.prof(event='dvm_fail', uid=cfg['pid'])
                    raise Exception("DVM process disappeared")


        # ----------------------------------------------------------------------
        def _watch_dvm():

            log.info('starting DVM watcher')

            retval = dvm_process.poll()
            while retval is None:
                line = dvm_process.stdout.readline().strip()
                if line:
                    log.debug('dvm output: %s', line)
                else:
                    time.sleep(1.0)

            if retval != 0:
                # send a kill signal to the main thread.
                # We know that Python and threading are likely not to play well
                # with signals - but this is an exceptional case, and not part
                # of the stadard termination sequence.  If the signal is
                # swallowed, the next `orte-submit` call will trigger
                # termination anyway.
                os.kill(os.getpid())

            log.info('DVM stopped (%d)' % dvm_process.returncode)
        # ----------------------------------------------------------------------

        dvm_watcher = mt.Thread(target=_watch_dvm)
        dvm_watcher.daemon = True
        dvm_watcher.start()

        lm_info = {'dvm_uri'     : dvm_uri,
                   'version_info': {name: orte_info}}

        # we need to inform the actual LaunchMethod instance about the DVM URI.
        # So we pass it back to the ResourceManager which will keep it in an
        # 'lm_info', which will then be passed as part of the slots via the
        # scheduler
        return lm_info


    # --------------------------------------------------------------------------
    #
    @classmethod
    def rm_shutdown_hook(cls, name, cfg, rm, lm_info, log, profiler):
        """
        This hook is symmetric to the config hook above, and is called during
        shutdown sequence, for the sake of freeing allocated resources.
        """

        if 'dvm_uri' in lm_info:
            try:
                log.info('terminating dvm')
                orterun = ru.which('orterun')
                if not orterun:
                    raise Exception("Couldn't find orterun")
                ru.sh_callout('%s --hnp %s --terminate'
                             % (orterun, lm_info['dvm_uri']))
                profiler.prof(event='dvm_stop', uid=cfg['pid'])

            except Exception as e:
                # use the same event name as for runtime failures - those are
                # not distinguishable at the moment from termination failures
                profiler.prof(event='dvm_fail', uid=cfg['pid'], msg=e)
                log.exception('dvm termination failed')


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        self.launch_command = ru.which('orterun')


    # --------------------------------------------------------------------------
    #
    def construct_command(self, cu, launch_script_hop):

        slots        = cu['slots']
        cud          = cu['description']
        task_exec    = cud['executable']
        task_mpi     = bool('mpi' in cud.get('cpu_process_type', '').lower())
        task_cores   = cud.get('cpu_processes', 0) + cud.get('gpu_processes', 0)
        task_env     = cud.get('environment') or dict()
        task_args    = cud.get('arguments')   or list()
        task_argstr  = self._create_arg_string(task_args)

        self._log.debug('prep %s', cu['uid'])

        if 'lm_info' not in slots:
            raise RuntimeError('No lm_info to launch via %s: %s'
                               % (self.name, slots))

        if not slots['lm_info']:
            raise RuntimeError('lm_info missing for %s: %s'
                               % (self.name, slots))

        if 'dvm_uri' not in slots['lm_info']:
            raise RuntimeError('dvm_uri not in lm_info for %s: %s'
                               % (self.name, slots))

        dvm_uri = slots['lm_info']['dvm_uri']

        if task_argstr: task_command = "%s %s" % (task_exec, task_argstr)
        else          : task_command = task_exec

        env_string = ''
        env_list   = self.EXPORT_ENV_VARIABLES + list(task_env.keys())
        if env_list:
            for var in env_list:
                env_string += '-x "%s" ' % var

        # Construct the hosts_string, env vars
        hosts_string = ''
        depths       = set()
        for node in slots['nodes']:

            # add all cpu and gpu process slots to the node list.
            for _        in node['core_map']: hosts_string += '%s,' % node['uid']
            for _        in node['gpu_map' ]: hosts_string += '%s,' % node['uid']
            for cpu_slot in node['core_map']: depths.add(len(cpu_slot))

        # assert(len(depths) == 1), depths
        # depth = list(depths)[0]

        # FIXME: is this binding correct?
      # if depth > 1: map_flag = '--bind-to none --map-by ppr:%d:core' % depth
      # else        : map_flag = '--bind-to none'
        map_flag = '--bind-to none'

        # remove trailing ','
        hosts_string = hosts_string.rstrip(',')

        # Additional (debug) arguments to orterun
        if os.environ.get('RADICAL_PILOT_ORTE_VERBOSE'):
            debug_strings = ['-display-devel-map',
                             '-display-allocation',
                             '--debug-devel',
                             '--mca oob_base_verbose 100',
                             '--mca rml_base_verbose 100'
                            ]
        else:
            debug_strings = []
        debug_string = ' '.join(debug_strings)

        if task_mpi: np_flag = '-np %s' % task_cores
        else       : np_flag = '-np 1'

        command = '%s %s --hnp "%s" %s %s -host %s %s %s' % (
                  self.launch_command, debug_string, dvm_uri, np_flag,
                  map_flag, hosts_string, env_string, task_command)

        return command, None


# ------------------------------------------------------------------------------





__author__    = 'RADICAL-Cybertools Team'
__email__     = 'info@radical-cybertools.org'
__copyright__ = 'Copyright 2020, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import logging
import os
import signal
import time

import subprocess    as mp
import threading     as mt

import radical.utils as ru

from .base import LaunchMethod


# ------------------------------------------------------------------------------
#
class PRTE2(LaunchMethod):

    """
    PMIx Reference Run Time Environment v2
    (https://github.com/openpmix/prrte)
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, cfg, session):

        LaunchMethod.__init__(self, name, cfg, session)

        # We remove all PRUN related environment variables from the launcher
        # environment, so that we can use PRUN for both launch of the
        # (sub-)agent and CU execution.
        self.env_removables.extend(['OMPI_', 'OPAL_', 'PMIX_'])

        self._verbose = bool(os.environ.get('RADICAL_PILOT_PRUN_VERBOSE'))

    # --------------------------------------------------------------------------
    #
    def _configure(self):

        if ru.which('prun'):
            self.launch_command = 'prun'
        else:
            raise Exception('prun not found')

    # --------------------------------------------------------------------------
    #
    @classmethod
    def rm_config_hook(cls, name, cfg, rm, log, profiler):

        prte = ru.which('prte')
        if not prte:
            raise Exception('prte not found')

        # get OpenRTE/PRTE version
        prte_info = {}
        out, _, _ = ru.sh_callout('prte_info | grep "RTE"', shell=True)
        for line in out.split('\n'):
            line = line.strip()

            if not prte_info.get('name'):
                if 'Open RTE' in line:
                    prte_info['name'] = 'Open RTE'
                elif 'PRTE' in line:
                    prte_info['name'] = 'PRTE'

            if 'RTE:' in line:
                prte_info['version'] = line.split(':')[1].strip()
            elif 'RTE repo revision:' in line:
                prte_info['version_detail'] = line.split(':')[1].strip()

        if prte_info.get('name'):
            log.info('version of %s: %s [%s]',
                     prte_info['name'],
                     prte_info.get('version'),
                     prte_info.get('version_detail'))
        else:
            log.info('version of OpenRTE/PRTE not determined')

        # write hosts file
        furi    = '%s/prrte.uri'   % os.getcwd()
        fhosts  = '%s/prrte.hosts' % os.getcwd()
        vm_size = len(rm.node_list)

        with open(fhosts, 'w') as f:
            for node in rm.node_list:
                f.write('%s slots=%d\n' % (node[0], rm.cores_per_node * rm.smt))

        prte += ' --prefix %s'     % os.environ.get('PRRTE_PREFIX', '')
        prte += ' --report-uri %s' % furi
        prte += ' --hostfile %s'   % fhosts

        if profiler.enabled:  # prte profiling
            # version 1 vs 2
            # prte += ' --pmca orte_state_base_verbose 1'
            prte += ' --mca orte_state_base_verbose 1'

        # large tasks imply large message sizes and we need to account for that
        _base_max_msg_size = 1024 * 1024 * 1024 * 1
        # FIXME: we should derive the message size from DVM size - smaller DVMs
        #        will never need large messages, as they can't run large tasks

        # version 1 vs 2
        # prte += ' --pmca ptl_base_max_msg_size %d' % _base_max_msg_size
        # # prte += ' --pmca rmaps_base_verbose 5'
        prte += ' --mca ptl_base_max_msg_size %d' % _base_max_msg_size

        # debug mapper problems for large tasks
        if log.isEnabledFor(logging.DEBUG):
            # version 1 vs 2
            # prte += ' -pmca orte_rmaps_base_verbose 100'
            prte += ' --mca orte_rmaps_base_verbose 100'

        # 2 temporary tweaks on Summit (not needed for the long run)
        # (1) avoid 64 node limit (ssh connection limit)
        # prte += ' --pmca plm_rsh_no_tree_spawn 1'
        # (2) ensure 1 ssh per dvm
        # prte += ' --pmca plm_rsh_num_concurrent %d' % vm_size

        # version 2
        prte += ' --mca plm_rsh_no_tree_spawn 1'
        prte += ' --mca plm_rsh_num_concurrent %d' % vm_size
        # FIXME: these 2 options above might not be needed
        #        (corresponding experiments should be conducted)

        # Use (g)stdbuf to disable buffering.  We need this to get the
        # "DVM ready" message to ensure DVM startup completion
        #
        # The command seems to be generally available on Cray's,
        # if not, we can code some home-cooked pty stuff (TODO)
        stdbuf_cmd = ru.which(['stdbuf', 'gstdbuf'])
        if not stdbuf_cmd:
            raise Exception('(g)stdbuf not found')
        stdbuf_arg = '-oL'

        # base command = (g)stdbuf <args> + prte + prte_args + debug_args
        cmd = '%s %s %s ' % (stdbuf_cmd, stdbuf_arg, prte)

        # additional (debug) arguments to prte
        verbose = bool(os.environ.get('RADICAL_PILOT_PRUN_VERBOSE'))
        if verbose:
            cmd += ' '.join(['--debug-devel',
                             '--mca odls_base_verbose 100',
                             '--mca rml_base_verbose 100'])

        cmd = cmd.strip()

        log.info('start prte on %d nodes [%s]', vm_size, cmd)
        profiler.prof(event='dvm_start', uid=cfg['pid'])

        dvm_process = mp.Popen(cmd.split(), stdout=mp.PIPE, stderr=mp.STDOUT)

        # ----------------------------------------------------------------------
        def _watch_dvm():

            log.info('starting prte watcher')

            retval = dvm_process.poll()
            while retval is None:
                line = dvm_process.stdout.readline().strip()
                if line:
                    log.debug('prte output: %s', line)
                else:
                    time.sleep(1.0)

            if retval != 0:
                # send a kill signal to the main thread.
                # We know that Python and threading are likely not to play well
                # with signals - but this is an exceptional case, and not part
                # of the standard termination sequence.  If the signal is
                # swallowed, the next `prun` call will trigger
                # termination anyway.
                os.kill(os.getpid(), signal.SIGKILL)
                raise RuntimeError('PRTE DVM died')

            log.info('prte stopped (%d)', dvm_process.returncode)
        # ----------------------------------------------------------------------

        dvm_watcher = mt.Thread(target=_watch_dvm)
        dvm_watcher.daemon = True
        dvm_watcher.start()

        dvm_uri = None
        for _ in range(100):

            time.sleep(0.5)
            try:
                with open(furi, 'r') as fin:
                    for line in fin.readlines():
                        if '://' in line:
                            dvm_uri = line.strip()
                            break

            except Exception as e:
                log.debug('DVM check - uri file missing: %s...', str(e)[:24])
                time.sleep(0.5)

            if dvm_uri:
                break

        if not dvm_uri:
            raise Exception('DVMURI not found')

        log.info('prte startup successful: [%s]', dvm_uri)

        # in some cases, the DVM seems to need some additional time to settle.
        # FIXME: this should not be needed
        time.sleep(10)
        profiler.prof(event='dvm_ok', uid=cfg['pid'])

        lm_info = {'dvm_uri'     : dvm_uri,
                   'version_info': prte_info,
                   'cvd_id_mode' : 'physical'}

        # we need to inform the actual LaunchMethod instance about the prte URI.
        # So we pass it back to the ResourceManager which will keep it in an
        # 'lm_info', which will then be passed as part of the slots via the
        # scheduler
        return lm_info

    # --------------------------------------------------------------------------
    #
    @classmethod
    def rm_shutdown_hook(cls, name, cfg, rm, lm_info, log, profiler):
        """
        This hook is symmetric to the hook `rm_config_hook`, it is called
        during shutdown sequence, for the sake of freeing allocated resources.
        """
        if lm_info.get('dvm_uri'):
            try:
                log.info('terminating prte')
                prun = ru.which('prun')
                if not prun:
                    raise Exception('prun not found')
                ru.sh_callout('%s --dvm-uri "%s" --terminate' %
                              (prun, lm_info['dvm_uri']))
                profiler.prof(event='dvm_stop', uid=cfg['pid'])

            except Exception as e:
                # use the same event name as for runtime failures - those are
                # not distinguishable at the moment from termination failures
                profiler.prof(event='dvm_fail', uid=cfg['pid'], msg=e)
                log.exception('prte termination failed')

    # --------------------------------------------------------------------------
    #
    def construct_command(self, cu, launch_script_hop):

        time.sleep(0.1)

        slots     = cu['slots']
        cud       = cu['description']

        task_exec = cud['executable']
        task_args = cud.get('arguments') or []
        task_env  = cud.get('environment') or {}

        n_procs   = cud.get('cpu_processes') or 1
        n_threads = cud.get('cpu_threads') or 1

        if not slots.get('lm_info'):
            raise RuntimeError('lm_info not set (%s): %s' % (self.name, slots))

        if not slots['lm_info'].get('dvm_uri'):
            raise RuntimeError('dvm_uri not set (%s): %s' % (self.name, slots))

        dvm_uri = '--dvm-uri "%s"' % slots['lm_info']['dvm_uri']

        # version 1 vs 2
        # flags  = ' -np %d --cpus-per-proc %d' % (n_procs, n_threads)
        # flags += ' --bind-to hwthread:overload-allowed --use-hwthread-cpus'
        # flags += ' --oversubscribe'
        flags  = ' -np %d --map-by :PE=%d' % (n_procs, n_threads)
        flags += ' --bind-to hwthread:overload-allowed'

        # DVM startup
        _base_max_msg_size = 1024 * 1024 * 1024 * 1

        # version 1 vs 2
        # flags += ' --pmca ptl_base_max_msg_size %d' % _base_max_msg_size
        # # flags += ' --pmca rmaps_base_verbose 5'
        flags += ' --mca ptl_base_max_msg_size %d' % _base_max_msg_size

        if 'nodes' not in slots:
            # this task is unscheduled - we leave it to PRRTE/PMI-X
            # to correctly place the task
            pass

        else:
            # FIXME: ensure correct binding for procs and threads via slotfile

            # enact the scheduler's host placement.  For now, we leave socket,
            # core and thread placement to the prted, and just add all process
            # slots to the host list
            hosts = ','.join([node['name'] for node in slots['nodes']])
            flags += ' --host %s' % hosts

        # additional (debug) arguments to prun
        debug_flags = '--verbose'  # needed to get prte profile events
        if self._verbose:
            debug_flags += ' ' + ' '.join(['--report-bindings'])
                                         # '--debug-devel',
                                         # '--display-devel-map',
                                         # '--display-allocation'])

        env_list = self.EXPORT_ENV_VARIABLES + list(task_env.keys())
        if env_list:
            envs = ' '.join(['-x "%s"' % k for k in env_list])
        else:
            envs = ''

        task_args_str = self._create_arg_string(task_args)
        if task_args_str:
            task_exec += ' %s' % task_args_str

        cmd = '%s %s %s %s %s %s' % (
            self.launch_command, dvm_uri, flags, debug_flags, envs, task_exec)

        return cmd, None

# ------------------------------------------------------------------------------




__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import radical.utils as ru

from .base import LaunchMethod


# ------------------------------------------------------------------------------
#
class MPIRun(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, cfg, session):

        self._mpt    = False
        self._rsh    = False
        self._ccmrun = ''
        self._dplace = ''

        LaunchMethod.__init__(self, name, cfg, session)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        if '_rsh' in self.name.lower():
            self._rsh = True
            self.launch_command = ru.which(['mpirun_rsh',  # Gordon (SDSC)
                                            'mpirun'       # general case
                                           ])

        elif '_mpt' in self.name.lower():
            self._mpt = True
            self.launch_command = ru.which(['mpirun_mpt',  # Cheyenne (NCAR)
                                            'mpirun'       # general case
                                           ])
        else:
            self.launch_command = ru.which(['mpirun-mpich-mp',    # Mac OSX
                                            'mpirun-openmpi-mp',  # Mac OSX
                                            'mpirun',             # general case
                                           ])

        # cheyenne is special: it needs MPT behavior (no -host) even for the
        # default mpirun (not mpirun_mpt).
        if 'cheyenne' in self._cfg.resource.lower():
            self._mpt = True

        # don't use the full pathname as the user might load a different
        # compiler / MPI library suite from his CU pre_exec that requires
        # the launcher from that version -- see #572.
        # FIXME: then why are we doing this LM setup in the first place??
        if self.launch_command:
            self.launch_command = os.path.basename(self.launch_command)


        # do we need ccmrun or dplace?
        if '_ccmrun' in self.name:
            self._ccmrun = ru.which('ccmrun')
            if not self._ccmrun:
                raise RuntimeError("ccmrun not found!")

        if '_dplace' in self.name:
            self._dplace = ru.which('dplace')
            if not self._dplace:
                raise RuntimeError("dplace not found!")

        self.mpi_version, self.mpi_flavor = \
                                       self._get_mpi_info(self.launch_command)


    # --------------------------------------------------------------------------
    #
    def construct_command(self, cu, launch_script_hop):

        slots        = cu['slots']
        uid          = cu['uid']
        cud          = cu['description']
        sandbox      = cu['unit_sandbox_path']
        task_exec    = cud['executable']
        task_threads = cud.get('cpu_threads', 1)
        task_env     = cud.get('environment') or dict()
        task_args    = cud.get('arguments')   or list()
        task_argstr  = self._create_arg_string(task_args)

        if '_dplace' in self.name and task_threads > 1:
            # dplace pinning would disallow threads to map to other cores
            raise ValueError('dplace can not place threads [%d]' % task_threads)

        # Construct the executable and arguments
        if task_argstr: task_command = "%s %s" % (task_exec, task_argstr)
        else          : task_command = task_exec

        env_string = ''
        env_list   = self.EXPORT_ENV_VARIABLES + list(task_env.keys())
        if env_list:

            if self.mpi_flavor == self.MPI_FLAVOR_HYDRA:
                env_string = '-envlist "%s"' % ','.join(env_list)

            elif self.mpi_flavor == self.MPI_FLAVOR_OMPI:
                for var in env_list:
                    env_string += '-x "%s" ' % var

        # Cheyenne is the only machine that requires mpirun_mpt.  We then
        # have to set MPI_SHEPHERD=true
        if self._mpt:
            if not cu['description'].get('environment'):
                cu['description']['environment'] = dict()
            cu['description']['environment']['MPI_SHEPHERD'] = 'true'

        # Extract all the hosts from the slots
        host_list = list()
        core_list = list()
        save_list = list()

        for node in slots['nodes']:

            for cpu_proc in node['core_map']:
                host_list.append(node['name'])
                core_list.append(cpu_proc[0])
                # FIXME: inform this proc about the GPU to be used

            if '_dplace' in self.name and save_list:
                assert(save_list == core_list), 'inhomog. core sets (dplace)'
            else:
                save_list = core_list

        if '_dplace' in self.name:
            self._dplace += ' -c '
            self._dplace += ','.join(core_list)


        # If we have a CU with many cores, we will create a hostfile and pass
        # that as an argument instead of the individual hosts
        hosts_string     = ''
        mpt_hosts_string = ''
        if len(host_list) > 42:

            # Create a hostfile from the list of hosts
            hostfile = self._create_hostfile(sandbox, uid, host_list,
                                             impaired=True)
            if self._mpt: hosts_string = '-file %s'     % hostfile
            else        : hosts_string = '-hostfile %s' % hostfile

        else:
            # Construct the hosts_string ('h1,h2,..,hN')
            if self._mpt: mpt_hosts_string = '%s'       % ",".join(host_list)
            else        : hosts_string     = '-host %s' % ",".join(host_list)

        # -np:  usually len(host_list), meaning N processes over N hosts, but
        # for Cheyenne (mpt) the specification of -host lands N processes on
        # EACH host, where N is specified as arg to -np
        if self._mpt: np = 1
        else        : np = len(host_list)

        command = ("%s %s %s -np %d %s %s %s %s" %
                   (self._ccmrun, self.launch_command, mpt_hosts_string,
                    np, self._dplace, hosts_string, env_string,
                    task_command)).strip()

        return command, None


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import time

import threading     as mt

import radical.utils as ru

from ...  import states    as rps
from ...  import constants as rpc

from .base import AgentExecutingComponent


# ------------------------------------------------------------------------------
#
class Sleep(AgentExecutingComponent) :

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        AgentExecutingComponent.__init__ (self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        self._pwd = os.getcwd()

        self.register_input(rps.AGENT_EXECUTING_PENDING,
                            rpc.AGENT_EXECUTING_QUEUE, self.work)

        self.register_output(rps.AGENT_STAGING_OUTPUT_PENDING,
                             rpc.AGENT_STAGING_OUTPUT_QUEUE)

        self.register_publisher (rpc.AGENT_UNSCHEDULE_PUBSUB)

        self._terminate  = mt.Event()
        self._tasks_lock = ru.RLock()
        self._tasks      = list()
        self._delay      = 0.1

        self._watcher = mt.Thread(target=self._timed)
        self._watcher.daemon = True
        self._watcher.start()


    # --------------------------------------------------------------------------
    #
    def finalize(self):

        self._terminate.set()
        self._watcher.join()


    # --------------------------------------------------------------------------
    #
    def work(self, units):

        if not isinstance(units, list):
            units = [units]

        self.advance(units, rps.AGENT_EXECUTING, publish=True, push=False)

        now = time.time()
        for t in units:
          # assert(t['description']['executable'].endswith('sleep'))
            t['to_finish'] = now + float(t['description']['arguments'][0])

        for t in units:
            uid = t['uid']
            self._prof.prof('exec_start',    uid=uid)
            self._prof.prof('exec_ok',       uid=uid)
            self._prof.prof('cu_start',      uid=uid)
            self._prof.prof('cu_exec_start', uid=uid)
            self._prof.prof('app_start',     uid=uid)

        with self._tasks_lock:
            self._tasks.extend(units)


    # --------------------------------------------------------------------------
    #
    def _timed(self):

        while not self._terminate.is_set():

            time.sleep(self._delay)

            with self._tasks_lock:
                now = time.time()
                to_finish   = [t for t in self._tasks if t['to_finish'] <= now]
                self._tasks = [t for t in self._tasks if t['to_finish'] >  now]

            for t in to_finish:
                uid = t['uid']
                t['target_state'] = 'DONE'
                self._prof.prof('app_stop',         uid=uid)
                self._prof.prof('cu_exec_stop',     uid=uid)
                self._prof.prof('cu_stop',          uid=uid)
                self._prof.prof('exec_stop',        uid=uid)
                self._prof.prof('unschedule_start', uid=uid)
                self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, t)

            self.advance(to_finish, rps.AGENT_STAGING_OUTPUT_PENDING,
                                    publish=True, push=True)


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import radical.saga            as rs
import radical.saga.filesystem as rsfs
import radical.utils           as ru

from .. import utils           as rpu
from .. import constants       as rpc
from .. import states          as rps


# ------------------------------------------------------------------------------
#
class Stager(rpu.Worker):
    '''
    A Stager will receive staging requests, perform the respective staging
    actions, and will respond with a completion message.  At the moment, the
    stager uses the CONTROL pubsub to communicate with components which require
    its staging capabilities.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        self._cache_lock    = ru.Lock()
        self._saga_fs_cache = dict()

        rpu.Worker.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        self._sid   = self._cfg['sid']
        self._dburl = self._cfg['dburl']

        self.register_input(rps.NEW, rpc.STAGER_REQUEST_QUEUE, self.work)
        self.register_publisher(rpc.STAGER_RESPONSE_PUBSUB)


    # --------------------------------------------------------------------------
    #
    @classmethod
    def create(cls, cfg, session):

        return cls(cfg, session)


    # --------------------------------------------------------------------------
    #
    def work(self, sds):
        '''
        receive staging requests, execute them, and reply with completion
        messages on the control channel
        '''

        self._log.debug('stager got  sds: %s ', [sd['uid'] for sd in sds])

        # handle the staging descriptions as bulk and report collective
        # completion.
        #
        # NOTE: should we allow bulk reshuffeling here?

        self._handle_staging(sds)

        # NOTE: we send completion notifications via pubsub, as we have no
        #       direct channel to whoever requested the trransfer.  That may
        #       raise scalability problems in the long run.
        self.publish(rpc.STAGER_RESPONSE_PUBSUB,
                     {'cmd': 'staging_result',
                      'arg': {'sds': sds}})
        self._log.debug('stager done sds: %s ', [sd['uid'] for sd in sds])

        return True


    # --------------------------------------------------------------------------
    #
    def _handle_staging(self, sds):

        # Iterate over all directives
        for sd in sds:

            # TODO: respect flags in directive

            action  = sd['action']
            flags   = sd['flags']
            uid     = sd['uid']
            src     = sd['source']
            tgt     = sd['target']
            prof_id = sd.get('prof_id')   # staging on behalf of this entity

            assert(action in [rpc.COPY, rpc.LINK, rpc.MOVE, rpc.TRANSFER])

            self._prof.prof('staging_start', uid=prof_id, msg=uid)

            if action in [rpc.COPY, rpc.LINK, rpc.MOVE]:
                self._prof.prof('staging_fail', uid=prof_id, msg=uid)
                raise ValueError("invalid action '%s' on stager" % action)

            self._log.info('transfer %s', src)
            self._log.info('      to %s', tgt)

            # open the staging directory for the target, and cache it
            # url used for cache: tgt url w/o path
            tmp      = rs.Url(tgt)
            tmp.path = '/'
            key      = str(tmp)

            with self._cache_lock:
                if key in self._saga_fs_cache:
                    fs = self._saga_fs_cache[key]

                else:
                    fs = rsfs.Directory(key, session=self._session)
                    self._saga_fs_cache[key] = fs

            fs.copy(src, tgt, flags=flags)

            sd['state'] = rps.DONE

            self._prof.prof('staging_stop', uid=prof_id, msg=uid)


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import signal

import radical.utils as ru

from .base import ResourceManager


# ------------------------------------------------------------------------------
#
class SGE(ResourceManager):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        ResourceManager.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        sge_hostfile = os.environ.get('PE_HOSTFILE')
        if sge_hostfile is None:
            msg = "$PE_HOSTFILE not set!"
            self._log.error(msg)
            raise RuntimeError(msg)

        def _sigusr2_handler():

            self._log.warn('caught sigusr2')
            self.stop()

        signal.signal(signal.SIGUSR1, _sigusr2_handler)

        # SGE core configuration might be different than what multiprocessing
        # announces
        # Alternative: "qconf -sq all.q|awk '/^slots *[0-9]+$/{print $2}'"

        # Parse SGE hostfile for nodes
        sge_node_list = [line.split()[0] for line in open(sge_hostfile)]
        self._log.info("Found PE_HOSTFILE %s. Expanded to: %s", sge_hostfile, sge_node_list)

        # Parse SGE hostfile for cores
        sge_cores_count_list = [int(line.split()[1]) for line in open(sge_hostfile)]
        sge_core_counts      = list(set(sge_cores_count_list))
        sge_gpus_per_node    = self._cfg.get('gpus_per_node', 0)  # FIXME GPU

        sge_lfs_per_node     = {'path' : ru.expand_env(
                                           self._cfg.get('lfs_path_per_node')),
                               'size' :    self._cfg.get('lfs_size_per_node', 0)
                              }

        # Check if nodes have the same core count
        if len(sge_core_counts) == 1:
            sge_cores_per_node = min(sge_core_counts)
            self._log.info("Found unique core counts: %s Using: %d", sge_core_counts, sge_cores_per_node)

            # node names are unique, so can serve as node uids
            self.node_list      = [[node, node] for node in sge_node_list]
            self.cores_per_node = sge_cores_per_node
            self.gpus_per_node  = sge_gpus_per_node
            self.lfs_per_node   = sge_lfs_per_node

        else:
            # In case of non-homogeneous counts, consider all slots be single core
            sge_cores_per_node = 1
            self._log.info("Found unique core counts: %s Using: %d", sge_core_counts, sge_cores_per_node)
            self.cores_per_node = sge_cores_per_node
            self.gpus_per_node  = sge_gpus_per_node
            self.lfs_per_node   = sge_lfs_per_node

            # Expand node list, create unique IDs for each core
            self.node_list = []
            for node, cores in zip(sge_node_list, sge_cores_count_list):
                for core in cores:
                    self.node_list.append(node, '%s_%s' % (node, core))


# ------------------------------------------------------------------------------





import os
import sys
import time
import datetime

import radical.utils as ru

from ..states import AGENT_EXECUTING


_CACHE_BASEDIR = '/tmp/rp_cache_%d/' % os.getuid ()


# ------------------------------------------------------------------------------
#
def bson2json (bson_data):

    # thanks to
    # http://stackoverflow.com/questions/16586180/ \
    #                          typeerror-objectid-is-not-json-serializable

    import json
    from   bson.objectid import ObjectId

    class MyJSONEncoder (json.JSONEncoder):
        def default (self, o):
            if  isinstance (o, ObjectId):
                return str (o)
            if  isinstance (o, datetime.datetime):
                seconds  = time.mktime (o.timetuple ())
                seconds += (o.microsecond / 1000000.0)
                return seconds
            return json.JSONEncoder.default (self, o)

    return ru.parse_json (MyJSONEncoder ().encode (bson_data))


# ------------------------------------------------------------------------------
#
def get_session_ids(db):

    # this is not bein cashed, as the session list can and will change freqently
    return db.collection_names(include_system_collections=False)


# ------------------------------------------------------------------------------
def get_last_session(db):

    # this assumes that sessions are ordered by time -- which is the case at
    # this point...
    return get_session_ids(db)[-1]


# ------------------------------------------------------------------------------
def get_session_docs(db, sid, cache=None, cachedir=None):

    # session docs may have been cached in /tmp/rp_cache_<uid>/<sid>.json -- in
    # that case we pull it from there instead of the database, which will be
    # much quicker.  Also, we do cache any retrieved docs to that place, for
    # later use.  An optional cachdir parameter changes that default location
    # for lookup and storage.
    if  not cachedir:
        cachedir = _CACHE_BASEDIR

    if  not cache:
        cache = "%s/%s.json" % (cachedir, sid)

    try:
        if  os.path.isfile (cache):
          # print 'using cache: %s' % cache
            return ru.read_json (cache)
    except Exception as e:
        # continue w/o cache
        sys.stderr.write("cannot read session cache at %s (%s)\n" % (cache, e))


    # cache not used or not found -- go to db
    json_data = dict()

    # convert bson to json, i.e. serialize the ObjectIDs into strings.
    json_data['session'] = bson2json(list(db[sid].find({'type': 'session'})))
    json_data['pmgr'   ] = bson2json(list(db[sid].find({'type': 'pmgr'   })))
    json_data['pilot'  ] = bson2json(list(db[sid].find({'type': 'pilot'  })))
    json_data['umgr'   ] = bson2json(list(db[sid].find({'type': 'umgr'   })))
    json_data['unit'   ] = bson2json(list(db[sid].find({'type': 'unit'   })))

    if  len(json_data['session']) == 0:
        raise ValueError ('no session %s in db' % sid)

  # if  len(json_data['session']) > 1:
  #     print 'more than one session document -- picking first one'

    # there can only be one session, not a list of one
    json_data['session'] = json_data['session'][0]

    # we want to add a list of handled units to each pilot doc
    for pilot in json_data['pilot']:

        pilot['unit_ids'] = list()

        for unit in json_data['unit']:

            if  unit['pilot'] == pilot['uid']:
                pilot['unit_ids'].append (unit['uid'])

    # if we got here, we did not find a cached version -- thus add this dataset
    # to the cache
    try:
        os.system ('mkdir -p %s' % cachedir)
        ru.write_json (json_data, "%s/%s.json" % (cachedir, sid))
    except Exception:
        # we can live without cache, no problem...
        pass

    return json_data


# ------------------------------------------------------------------------------
#
def get_session_slothist(db, sid, cache=None, cachedir=None):
    """
    For all pilots in the session, get the slot lists and slot histories. and
    return as list of tuples like:

         [pid   ,     [     [host,   slot]],     [      [slotstate, timestamp]]]
    tuple(string, list(tuple(string, int )), list(tuple (string   , datetime )))
    """

    docs = get_session_docs(db, sid, cache, cachedir)

    ret = dict()

    for pilot_doc in docs['pilot']:

        pid          = pilot_doc['uid']
        slot_names   = list()
        slot_infos   = dict()
        slot_started = dict()

        nodes   = pilot_doc['nodes']
        n_cores = pilot_doc['cores_per_node']

        if  not nodes:
          # print "no nodes in pilot doc for %s" % pid
            continue

        for node in nodes:
            for core in range(n_cores):
                slot_name = "%s:%s" % (node, core)
                slot_names.append (slot_name)
                slot_infos  [slot_name] = list()
                slot_started[slot_name] = sys.maxsize

        for unit_doc in docs['unit']:
            if unit_doc['pilot'] == pilot_doc['uid']:

                started  = None
                finished = None
                for event in sorted (unit_doc['state_history'],
                                     key=lambda x: x['timestamp']):
                    if started:
                        finished = event['timestamp']
                        break
                    if event['state'] == AGENT_EXECUTING:
                        started = event['timestamp']

                if not started or not finished:
                  # print "no start/end for cu %s - ignored" % unit_doc['uid']
                    continue

                for slot_id in unit_doc['opaque_slots']:
                    if slot_id not in slot_infos:
                      # print "slot %s for pilot %s unknown - ignored" \
                      #     % (slot_id, pid)
                        continue

                    slot_infos[slot_id].append([started, finished])
                    slot_started[slot_id] = min(started, slot_started[slot_id])

        for slot_id in slot_infos:
            slot_infos[slot_id].sort(key=lambda x: float(x[0]))

        # we use the startup time to sort the slot names, as that gives a nicer
        # representation when plotting.  That sorting should probably move to
        # the plotting tools though... (FIXME)
        slot_names.sort (key=lambda x: slot_started[x])

        ret[pid] = dict()
        ret[pid]['started']    = pilot_doc['started']
        ret[pid]['finished']   = pilot_doc['finished']
        ret[pid]['slots']      = slot_names
        ret[pid]['slot_infos'] = slot_infos

    return ret


# ------------------------------------------------------------------------------
def get_session_events(db, sid, cache=None, cachedir=None):
    """
    For all entities in the session, create simple event tuples, and return
    them as a list

    [     [otype , oid   , pid   , timestamp, event name, odoc]]
    list(tuple(string, string, string, datetime , string    , dict))

    """

    docs = get_session_docs(db, sid, cache, cachedir)
    ret  = list()

    if  'session' in docs:
        doc   = docs['session']
        odoc  = dict()
        otype = 'session'
        oid   = doc['uid']
        ret.append(['state', otype, oid, None, doc['created'],
                    'created',   odoc])
        ret.append(['state', otype, oid, None, doc['connected'],
                    'connected', odoc])

    for doc in docs['pilot']:
        odoc  = dict()
        otype = 'pilot'
        oid   = doc['uid']

        for event in ['input_transfer_started',  'input_transfer_finished',
                      'output_transfer_started', 'output_transfer_finished']:
            if  event in doc:
                ret.append (['state', otype, oid, oid, doc[event], event, odoc])
            else:
                ret.append (['state', otype, oid, oid, None,       event, odoc])

        for event in doc['state_history']:
            ret.append (['state',     otype, oid, oid, event['timestamp'],
                         event['state'], odoc])

        if  'callbackhistory' in doc:
            for event in doc['callbackhistory']:
                ret.append (['callback',  otype, oid, oid, event['timestamp'],
                             event['state'], odoc])


    for doc in docs['unit']:
        odoc  = dict()
        otype = 'unit'
        oid   = doc['uid']
        pid   = doc['pilot']

        # TODO: change states to look for
        for event in ['input_transfer_started',  'input_transfer_finished',
                      'output_transfer_started', 'output_transfer_finished']:

            if  event in doc:
                ret.append (['state', otype, oid, pid, doc[event], event, doc])
            else:
                ret.append (['state', otype, oid, pid, None,       event, doc])

        for event in doc['state_history']:
            ret.append (['state',     otype, oid, pid, event['timestamp'],
                         event['state'], doc])

        # TODO: this probably needs to be "doc"
        if  'callbackhistory' in event:
            for event in doc['callbackhistory']:
                ret.append (['callback',  otype, oid, pid, event['timestamp'],
                             event['state'], doc])

    # we don't want None times, actually
    for r in list(ret):
        if  r[4] is None:
            ret.remove (r)

    ret.sort (key=lambda tup: tup[4])

    return ret


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__copyright__ = 'Copyright 2013-2016, http://radical.rutgers.edu'
__license__   = 'MIT'

import sys
import radical.utils as ru
import radical.pilot as rp

rpu = rp.utils


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    if len(sys.argv) <= 1:
        print("\n\tusage: %s <session_id>\n")
        sys.exit(1)

    sid = sys.argv[1]


    profiles = rpu.fetch_profiles(sid=sid, skip_existing=True)
    for p in profiles:
        print(p)

    profs = ru.read_profiles(profiles)

    for p in profs:
        print(type(p))

    prof = ru.combine_profiles(profs)

    print(len(prof))
    for entry in prof:
        print(entry)




# ------------------------------------------------------------------------------

